
/* XXX if 3rd 2401 responds, abort wind direction routine */
/* XXX if single device, use SKIP in place of MATCH */

#include "standard.h"
#include "crc.h"

extern int debug;

/* in ow.s */

bool ow_reset(void);
void
	ow_bits_in(byte *, int nbits),
	ow_bits_out(byte *, int nbits);
word
	ow_two_bits(void);

/* no more than 16 types fit into 'sortiment' variable */

enum types {
	DS1820, DS1822, /* DS18B20 = DS1822. keep these 1st */
	DS2405, DS2407, DS2408, DS2413, /* DS2406 = DS2407 */
	DS2423,
	DS2450, DS2436, DS2438,

	METATYPES,

	DS2409,
	DS2401,          /* peculiar handling, keep this last */
};

/*
 * one 2401 device gives just one bit value,
 * but one 2438 device gives 1 temperature
 * and 3 adc values. And 2408 creates whopping 8 input values.
 */
#define MAX_DEVICES  12
#define MAX_VALUES   16 /* bytes or bits, all the same. */

#define SEARCH_ROM     0xF0

#define READ_SCRATCH   0xBE
#define WRITE_SCRATCH  0x4E /* 2438 */
#define CONVERT_T      0x44 /* 1820 and 2438 */
#define CONVERT_V      0xB4 /* 2436 and 2438 */
#define CONVERT_AD     0x3C /* 2450 */
#define RECALL_MEMORY  0xB8 /* 2438 */

#define IDENT_SIZE 8

struct owdev {
	byte ident[IDENT_SIZE];
	byte type;

	/* bool (*unload)(struct owdev *); the switch(type) is just fine */

	/* struct owdev *parent; if hanging from ds2409 */

	union {
		struct {
			short celsius;   /* one temperature. half celsius everywhere */
		} s1820;

		struct {
			byte status;     /* one bit */
		} s2405;

		struct {
			byte infob;      /* one or two/three bits */
		} s2407;

		struct {
			byte status;     /* eight bits */
		} s2408;

		struct {
			byte status;     /* two bits, inputs in 0x5 */
		} s2413;

		struct {
			dword count[2];  /* two 32bit counters */
			stamp time;      /* of the read transfer */
		} s2423;

		struct {
			byte adc[4];     /* four A/D inputs */
		} s2450;

		struct {
			short celsius;   /* one temperature */
			byte  voltage;   /* one A/D input */
		} s2436;

		struct {
			short celsius;   /* one temperature */
			byte  adc[3];    /* three A/D inputs */
		} s2438;

		/* ds2401 and ds2409 are handled differently */
	} d;
};

STATIC struct owdev *end_devices, devices[MAX_DEVICES];

STATIC struct {
	void *where;
	union {
		word mask;
		word size;
	} how;
} ranking[MAX_VALUES];

STATIC word
	roundcnt,         /* toggle things in ow_round() if needed */
	sortiment,        /* what types there are, no more than 16 */
	byterank,         /* at what rank offset are byte values */
	accrank,          /* at what rank offset are counter values */
	numrank;          /* how many separate values we can provide */

STATIC short *temperature; /* randomly selected 1820, if any */

STATIC byte gident[IDENT_SIZE]; /* A. Dent */

#define any(type) (sortiment & (1 << (type)))

/* -------------------------------------------- */

word
ow_bitvalue(word rank)
{
	if (rank >= byterank) /* bits are ranked first */
		return (0);

	return (*(byte *) ranking[rank].where
	                & ranking[rank].how.mask);
}

word
ow_bytevalue(word rank)
{
	if ((rank += byterank) >= accrank)    /* bytes before counters */
		return (0);

	if (ranking[rank].how.size == 1)
		return (*(byte *) ranking[rank].where);
	else
		return (*(word *) ranking[rank].where); /* hmph */
}

dword
ow_accvalue(word rank)
{
	if ((rank += accrank) >= numrank)     /* counters are last */
		return (0);

	return (*(dword *) ranking[rank].where);
}

int
ow_fahrenheit()
{
	if (temperature) {
		short half_c = *temperature;

		/* some rounding craze */

		return ((half_c * 9 / 5 + 32 * 2 + 1) / 2);
	}
	return (0); /* hmph */
}

int
ow_celsius()
{
	if (temperature) {
		short half_c = *temperature;

		return (half_c >> 1);
	}
	return (0); /* hmph */
}

/* -------------------------------------------- */

#define RANK_VAL(what) rank_it(&(what), sizeof(what))

static void
rank_it(void *where, word mask)
{
	if (numrank < __SIZEOF(ranking)) {
		ranking[numrank].where    = where;
		ranking[numrank].how.mask = mask;
		numrank++;
	}
}

static void
rank_them()
{
	struct owdev *d;
	int i;

	temperature = NULL;

	/* bits */

	numrank = 0;

	for (d = devices; d < end_devices; ++d) {
		switch (d->type) {

		case DS2405:
			rank_it(&d->d.s2405.status, 0x01);
			break;

		case DS2407:
			rank_it(&d->d.s2407.infob, 0x04);     /* PIO A Sensed level */

			if (d->d.s2407.infob & 0x40) {
				rank_it(&d->d.s2407.infob, 0x08); /* PIO B Sensed level */
				rank_it(&d->d.s2407.infob, 0x80); /* SUPPLY is an input too */
			}
			break;

		case DS2408:
			for (i = 0; i < 8; ++i)
				rank_it(&d->d.s2408.status, byte_bitmask[i]);
			break;

		case DS2413:
			rank_it(&d->d.s2413.status, 1);
			rank_it(&d->d.s2413.status, 4);
			break;
		}
	}

	/* bytes */

	byterank = numrank;

	for (d = devices; d < end_devices; ++d) {
		switch (d->type) {

		case DS1820:
		case DS1822:
			temperature = &d->d.s1820.celsius; /* hmph */

			RANK_VAL(d->d.s1820.celsius);
			break;

		case DS2450:
			for (i = 0; i < 4; ++i)
				RANK_VAL(d->d.s2450.adc[i]);
			break;

		case DS2436:
			RANK_VAL(d->d.s2436.celsius);
			RANK_VAL(d->d.s2436.voltage);
			break;

		case DS2438:
			RANK_VAL(d->d.s2438.celsius);

			for (i = 0; i < 3; ++i)
				RANK_VAL(d->d.s2438.adc[i]);
			break;
		}
	}
	/* counter values */

	accrank = numrank;

	for (d = devices; d < end_devices; ++d) {
		switch (d->type) {

		case DS2423:
			for (i = 0; i < 2; ++i)
				RANK_VAL(d->d.s2423.count[i]);
			break;
		}
	}

	/* XXX rank word of 2401 existence now ? */
}

/* -------------------------------------------- */

static void
prident(byte *ident)
{
	int i;

	for (i = IDENT_SIZE - 1; --i > 0; ) /* skip family and crc bytes */
		print_byte(ident[i]);
}

static void
pr2407_channel(byte x)
{
	print_string((x & 0x04) ? " high" : " low");
	if ((x & 0x01) == 0)
		print_string(" pulldown");
	if (x & 0x10)
		print_string(" activity");
}

static void
prcelsius(short c)
{
	print_string(" ");
	print_signed_short(c >> 1);
	print_string(" C");
}

static void
prdevice(struct owdev *d)
{
	int i;

	prident(d->ident);
	print_string(":");

	switch (d->type) {

	case DS1820:
	case DS1822:
		prcelsius(d->d.s1820.celsius);
		break;

	case DS2405:
		print_string((d->d.s2405.status & 0x01) ? " high" : " low");
		break;

	case DS2407:
		pr2407_channel(d->d.s2407.infob);

		if (d->d.s2407.infob & 0x40) {
			print_string(";");
			pr2407_channel(d->d.s2407.infob >> 1);

			/* print the SUPPLY bit as another input */
			print_string(";");
			print_string((d->d.s2407.infob & 0x80) ? " high" : " low");
		}
		break;

	case DS2408:
		print_string(" 0x");
		print_byte(d->d.s2408.status);
		break;

	case DS2413:
		print_string((d->d.s2413.status & 1) ? " high" : " low");
		print_string((d->d.s2413.status & 4) ? " high" : " low");
		break;

	case DS2423:
		for (i = 0; i < 2; ++i) {
			print_string(" 0x");
			print_dword(&d->d.s2423.count[i]);
		}
		break;

	case DS2450:
		for (i = 0; i < 4; ++i) {
			print_string(" 0x");
			print_byte(d->d.s2450.adc[i]);
		}
		break;

	case DS2436:
		prcelsius(d->d.s2436.celsius);
		print_string(" 0x");
		print_byte(d->d.s2436.voltage);
		break;

	case DS2438:
		prcelsius(d->d.s2438.celsius);

		for (i = 0; i < 3; ++i) {
			print_string(" 0x");
			print_byte(d->d.s2438.adc[i]);
		}
		break;
	}
	eol();
}

/* -------------------------------------------- */

static void
bit_out(byte x)        /* LSbit of x */
{
	ow_bits_out(&x, 1);
}

static void
command(byte cmd)
{
	ow_reset();
	ow_bits_out(&cmd, 8);
}

static void
SKIP_xfer(byte *buf, int nout, int nin)
{
	command(0xCC); /* SKIP_ROM */

	if (nout)
		ow_bits_out(buf, nout * 8);
	if (nin)
		ow_bits_in(buf + nout, nin * 8);
}

static void
SKIP_cmd(byte cmd)
{
	SKIP_xfer(&cmd, 1, 0);
}

static void
MATCH_xfer(struct owdev *d, byte *buf, int nout, int nin)
{
	command(0x55); /* MATCH_ROM */

	ow_bits_out(d->ident, IDENT_SIZE * 8);

	if (nout)
		ow_bits_out(buf, nout * 8);
	if (nin)
		ow_bits_in(buf + nout, nin * 8);
}

static bool
SEARCH_access(byte *ident, word bytecnt)
{
	command(SEARCH_ROM);

	do {
		word bits = *ident++ | 0x100;

		do {
			/*
			 * looking for 0, must receive ?0 (0 first)
			 * looking for 1, must receive 0?
			 */
			if (ow_two_bits() & byte_bitmask[bits & 1])
				return (FALSE);

			bit_out(bits);
			bits >>= 1;
		} while (bits != 1);
	} while (--bytecnt);

	return (TRUE);
}

/* -------------------------------------------- */

/*
 *  This is messy:
 *
 *  Initialization (or re-i.) searches for any type,
 *  and all normal devices are added to 'devices'.
 *  Then, DS2401 are continuously scanned for,
 *  but they are never added to 'devices'.
 */
static void
device_found()
{
	char *name = NULL;
	byte type;

	switch (gident[0]) {
	case 0: case 255:     /* no hogging please */
		break;

	case 0x01:
		type = DS2401; name = "2401";
		break;

	case 0x05:
		type = DS2405; name = "2405";
		break;

	case 0x10:
		type = DS1820; name = "1820";
		break;

	case 0x12:
		type = DS2407; name = "2407"; /* or 2406, the same */
		break;

	case 0x1B:
		type = DS2436; name = "2436";
		break;

	case 0x1D:
		type = DS2423; name = "2423";
		break;

	case 0x1F:
		type = DS2409; name = "2409";
		break;

	case 0x20:
		type = DS2450; name = "2450";
		break;

	case 0x22:
		type = DS1822; name = "1822";
		break;

	case 0x26:
		type = DS2438; name = "2438";
		break;

	case 0x28:
		type = DS1822; name = "18B20"; /* lumped with DS1822 */
		break;

	case 0x29:
		type = DS2408; name = "2408";
		break;

	case 0x3A:
		type = DS2413; name = "2413";
		break;
	}
	if (debug) {
		prident(gident);
		print_string(" ");
		if (name) {
			print_string(name);
			if (end_devices >= &devices[__SIZEOF(devices)])
				print_string(" too many");
		} else {
			print_byte(gident[0]);
			print_string(" is unknown");
		}
		eol();
	}

	if (name) {
		if (type < METATYPES
		 && end_devices < &devices[__SIZEOF(devices)]) {
			end_devices->type = type;
			bcopy(gident, end_devices->ident, IDENT_SIZE);
			end_devices++;
		}
		sortiment |= word_bitmask[type];
	}
}

static void
select(byte i, byte bit)
{
	if (bit & 1)
		gident[i / 8] |= byte_bitmask[i & 7];
	else
		gident[i / 8] &= ~byte_bitmask[i & 7];

	bit_out(bit);
}

static bool
bitat(byte i)
{
	return ((gident[i / 8] & byte_bitmask[i & 7]) ? 1 : 0);
}

static void
redo_search(byte pos)
{
	byte i, x;

	command(SEARCH_ROM);

	/* "steer from memory" the select bits, ignore the responses */

	for (i = 0; i < pos; ++i) {
		x = (bitat(i) << 2) | 3;
		ow_bits_out(&x, 3);
	}
	x = 3;
	ow_bits_out(&x, 2);       /* 2 more to position to the fork */
}

static void
scan_bus(byte i)
{
	byte x;

	for ( ; i < IDENT_SIZE * 8; ++i) {
		x = ow_two_bits();
		if (x >= 3)
			return;           /* no one there, or even a short on bus */

		if (x == 0) {
			select(i, 0);
			scan_bus(i + 1);  /* handle zero-branch downstairs */
			redo_search(i);   /* and then continue in one-branch */
		}
		select(i, x != 2);    /* 1 unless a single zero */
	}

	if (ow_calc_crc8(gident, IDENT_SIZE) == 0)
		device_found();
}

/* ------------------------------------------------- */

static void
scan_2401()
{
	/* XXX connect them */

	gident[0] = 0x01; /* DS2401 family code, a byte or 8 bits */

	if (SEARCH_access(gident, 1)) /* search this family only */
		scan_bus(8);

	/* XXX disconnect them */
}

/*
 *  Assume that common case is 0x01 or 0x03 in some rotational position.
 *  Bad input causes 8 loops and return 360.
 */
int
wind_direction_from_bitmask(word bits)
{
	word deg = 0;

	do {
		if ((bits & ~2) == 1)
			break;
		bits = RRC_byte(bits);
		deg += 45;
	} while (deg < 360);

	if (bits == 3)
		deg += 22;

	return (deg);
}

/*
 *  8 switches in circle, those on one side pull a/d pin
 *  to ground and those on another bias it to mid voltage.
 *  both switches open, a/d pin is vcc.
 *  thresholds: less than 0x55, less than 0xAA and not.
 */
int
wind_direction_from_2450(byte *adc)
{
	word bits = 0x10; /* four shifts to 0x100 */

	do {
		bits <<= 1;

		if (*adc < 0x55)
			bits |= 0x10;
		else
		if (*adc < 0xAA)
			bits |= 0x01;

		adc++;
	} while ((bits & 0x100) == 0);

	return (wind_direction_from_bitmask(bits));
}

/* ------------------------------------------------- */

static bool
read_1820_temp(struct owdev *d)
{
	byte buf[1 + 9]; /* cmd / scratchpad,crc8 */
	short celsius;

	buf[0] = READ_SCRATCH;

	MATCH_xfer(d, buf, 1, 9);

	if (ow_calc_crc8(buf + 1, 9)) /* cmd is not included in crc */
		return (FALSE);

	/* resolution 0.5 celsius, 8+1 bits */

	celsius = ld_short_intel(buf + 1);

	if (d->type != DS1820)
		celsius >>= 3;        /* 1822 has 12+4 bits. drop excess */

	d->d.s1820.celsius = celsius;
	return (TRUE);
}

/* ------------------------------------------------- */

/* so contorted device */

static bool
read_2405_status(struct owdev *d)
{
	byte x;

	if (!SEARCH_access(d->ident, IDENT_SIZE)) /* "search" this dev only */
		return (FALSE);

	x = ow_two_bits();     /* read it 2 times, just for fun */
	if (x != 0 && x != 3)  /* should be the same bit all the time */
		return (FALSE);

	d->d.s2405.status = (x & 1);
	return (TRUE);
}

static bool
write_2405_control(struct owdev *d, word bit)
{
	/* XXX each MATCH_ROM ___toggles___ it */

	return (FALSE);
}

/* ------------------------------------------------- */

#if 0

static bool
read_2407_otprom(struct owdev *d)
{
	byte buf[3 + 128 + 2]; /* cmd,addr / data,crc16 */
	int i;

	buf[0] = 0xF0; /* 2407 Read Memory */
	buf[1] = 0x00;
	buf[2] = 0x00; /* read it all, 128 bytes starting from 0x0000 */

	MATCH_xfer(d, buf, 3, 128 + 2);

	if (ow_calc_crc16(buf, sizeof(buf)))
		return (FALSE);

	for (i = 0; i < 128; ++i) {
		if (debug) {
			print_byte(buf[3 + i]);
			if (i % 32 == 31)
				eol();
		}
	}
	return (TRUE);
}

#endif

static bool
read_2407_infob(struct owdev *d)
{
	byte buf[3 + 2 + 2]; /* cmd,ctrl1,2 / infobyte,pindata,crc16 */

	buf[0] = 0xF5; /* CHANNEL_ACCESS */
	buf[1] = 0xD5; /* Control 1: ALR, just read, CRC each byte */
	buf[2] = 0xFF; /* Control 2: all bits reserved */

	MATCH_xfer(d, buf, 3, 2 + 2);

	if (ow_calc_crc16(buf, sizeof(buf)))
		return (FALSE);

	d->d.s2407.infob = buf[3 + 0]; /* Channel Info Byte is all we need */

	return (TRUE);
}

static bool
write_2407_control(struct owdev *d, word bits)
{
	byte buf[4 + 2];

	bits &= 3;    /* two outputs, bit 0 for PIO-A, bit 1 for PIO-B */

	buf[0] = 0x55;  /* WRITE_STATUS */
	buf[1] = 0x07;
	buf[2] = 0x00;
	buf[3] = 0x1F | (bits << 5);

	MATCH_xfer(d, buf, 4, 2);

	return (ow_calc_crc16(buf, sizeof(buf)) == 0);
}

/* ------------------------------------------------- */

static bool
read_2408_registers(struct owdev *d)
{
	byte buf[3 + 8 + 2]; /* cmd / pins,latch,activity,2junk,csr,2junk */

	buf[0] = 0xF0; /* Read PIO Registers */
	buf[1] = 0x88;
	buf[2] = 0x00; /* located at 0x0088 to 0x0090 */

	MATCH_xfer(d, buf, 3, 8 + 2);

	if (ow_calc_crc16(buf, sizeof(buf)))
		return (FALSE);

	d->d.s2408.status = buf[3 + 0]; /* PIO Logic State */

	/* buf[3 + 1]; PIO Output   Latch State */
	/* buf[3 + 2]; PIO Activity Latch State */
	/* buf[3 + 5]; Control/Status */

	return (TRUE);
}

static bool
read_2408_pins(struct owdev *d)
{
	byte buf[3 + 2]; /* cmd / 0xAA,pins */
	byte x = 0xFF;   /* keep all outputs oc */

	buf[0] = 0x5A;     /* Channel Access Write */
	buf[1] = x;        /* output data */
	buf[2] = x ^ 0xFF; /* inverted output data */

	MATCH_xfer(d, buf, 3, 2);

	if (buf[3 + 0] != 0xAA) /* how silly can this be */
		return (FALSE);

	d->d.s2408.status = buf[3 + 1]; /* PIO Pin Status */

	return (TRUE);
}

/* --- looks much like the other device ... -------- */

static bool
read_2413_pins(struct owdev *d)
{
	byte buf[3 + 2]; /* cmd / 0xAA,status */
	byte x = 0xFF;   /* keep all (two LSbits) outputs oc */

	buf[0] = 0x5A;     /* PIO Access Write */
	buf[1] = x;        /* output data */
	buf[2] = x ^ 0xFF; /* inverted output data */

	MATCH_xfer(d, buf, 3, 2);

	/* 1st response byte: 0xAA constant */

	if (buf[3 + 0] != 0xAA)
		return (FALSE);

	/* 2nd: PIO Pin Status: B out, B in, A out, A in */

	x = buf[3 + 1] ^ 0xF0; /* upper nybble has the complement */

	if (((x >> 4) ^ x) & 0xF)
		return (FALSE);

	d->d.s2413.status = x;

	return (TRUE);
}

/* ------------------------------------------------- */

static bool
read_2423_counts(struct owdev *d)
{
	byte buf[3 + 1 + 8 + 2]; /* cmd / junk,count64,crc */
	int i, did = 0;

	for (i = 0; i < 2; ++i) {
		buf[0] = 0xA5;            /* READ_COUNTER */
		buf[1] = i ? 0xFF : 0xDF;
		buf[2] = 0x01;            /* pages 14 (1DF) and 15 (1FF) */

		MATCH_xfer(d, buf, 3, 1 + 8 + 2);

		now(&d->d.s2423.time);

		if (ow_calc_crc16(buf, sizeof(buf)) == 0) {
			++did;
			d->d.s2423.count[i] = ld_dword_intel(&buf[3 + 1]);
		}
	}
	return (did == 2);
}

/* ------------------------------------------------- */

/* XXX inefficient, but currently rarely used */

static bool
write_2450_byte(struct owdev *d, word addr, byte val)
{
	byte buf[4 + 2 + 1];  /* cmd / crc16,verify */

	buf[0] = 0x55;      /* WRITE_MEMORY */
	buf[1] = addr;
	buf[2] = addr >> 8; /* start address */
	buf[3] = val;

	MATCH_xfer(d, buf, 4, 2 + 1);

	if (ow_calc_crc16(buf, 4 + 2))
		return (FALSE);

	return (buf[4 + 2] == val);
}

static bool
config_2450(struct owdev *d)
{
	/* page 1 in addresses 0x08 to 0x0F inclusive */

	static byte pg1[] = {
		0x08, 0x01, /* A: not an output, 8bit a/d, no alarms, 5v range */
		0x08, 0x01, /* B: not an output, 8bit a/d, no alarms, 5v range */
		0x08, 0x01, /* C: not an output, 8bit a/d, no alarms, 5v range */
		0x08, 0x01, /* D: not an output, 8bit a/d, no alarms, 5v range */
	};
	int i;

	for (i = 0; i < sizeof(pg1); ++i)
		if (!write_2450_byte(d, 0x0008 + i, pg1[i]))
			return (FALSE);

	return (TRUE);
}

static bool
read_2450_adc(struct owdev *d)
{
	byte buf[3 + 8 + 2];

	buf[0] = 0xAA; /* READ_MEMORY */
	buf[1] = 0x00;
	buf[2] = 0x00; /* start address 0x0000 */

	MATCH_xfer(d, buf, 3, 8 + 2);

	if (ow_calc_crc16(buf, sizeof(buf)))
		return (FALSE);

	d->d.s2450.adc[0] = buf[3 + 1];
	d->d.s2450.adc[1] = buf[3 + 3];
	d->d.s2450.adc[2] = buf[3 + 5];
	d->d.s2450.adc[3] = buf[3 + 7];

	return (TRUE);
}

static bool
start_every_2450()
{
	byte buf[3 + 2];

	buf[0] = CONVERT_AD;
	buf[1] = 0x0F; /* input select mask: all of them */
	buf[2] = 0x55; /* read-out control: preload to zero */

	SKIP_xfer(buf, 3, 2); /* each one returns the same crc */

	return (ow_calc_crc16(buf, sizeof(buf)) == 0);
}


/* ------------------------------------------------- */

/*
 * too bad there is such a large gap
 *  reset,MATCH,id,2+24 bytes, about 18  msec
 *  reset,MATCH,id,2+2 bytes,reset,MATCH,id,2+2 bytes, about 15 msec
 */
static bool
read_2436_data(struct owdev *d)
{
	byte buf[2 + 2]; /* cmd,addr / data... */

	buf[0] = 0xB2; /* Read Registers */
	buf[1] = 0x60; /* Temp Word */

	MATCH_xfer(d, buf, 2, 2);

	/* drop excess resolution, to 0.5 celsius */

	d->d.s2436.celsius = ld_short_intel(&buf[2 + 0]) >> 7;

	buf[0] = 0xB2; /* Read Registers */
	buf[1] = 0x77; /* Voltage word (10 bits valid) */

	MATCH_xfer(d, buf, 2, 2);

	d->d.s2436.voltage = ld_word_intel(&buf[2 + 0]) >> 2;

	return (TRUE);
}

/* ------------------------------------------------- */

static void
config_every_2438()
{
	byte buf[2 + 1]; /* cmd,page / control_register */

	buf[0] = WRITE_SCRATCH;
	buf[1] = 0x00; /* page 0 */
	buf[2] = 0x01; /* AD=0 (Vad is input), EE=0, CA=0, IAD=1 */

	SKIP_xfer(buf, 2 + 1, 0);
}

static void
recall_every_2438()
{
	byte buf[2]; /* cmd,page */

	buf[0] = RECALL_MEMORY;
	buf[1] = 0x00;         /* page 0 */

	SKIP_xfer(buf, 2, 0);
}

static bool
read_2438_data(struct owdev *d)
{
	byte buf[2 + 8 + 1]; /* cmd,page / page_data,crc8 */

	buf[0] = READ_SCRATCH;
	buf[1] = 0x00;         /* page 0 */

	MATCH_xfer(d, buf, 2, 8 + 1);

	if (ow_calc_crc8(buf + 2, sizeof(buf) - 2)) /* cmd not included */
		return (FALSE);

	d->d.s2438.celsius = ld_short_intel(&buf[2 + 1]) >> 7;

	d->d.s2438.adc[0] = ld_word_intel(&buf[2 + 5]) >> 3; /* current */
	d->d.s2438.adc[1] = ld_word_intel(&buf[2 + 3]) >> 2; /* Vad */
	d->d.s2438.adc[2] = ~0;                              /* Vdd XXX */

	return (TRUE);
}

/* ------------------------------------------------- */

#if 0

static void
scan_2409_bus()
{
	byte buf[4];

	buf[0] = SMART_ON_MAIN;
	buf[1] = 0xFF;         /* reset stimulus */
	buf[2] = 0xFF;         /* reset response */
	buf[3] = SEARCH_ROM;

	MATCH_xfer(d, buf, 4, 0);
}

#endif

/* ------------------------------------------------- */

static bool
fondle(struct owdev *d)
{
	switch (d->type) {

	case DS1820:
	case DS1822:
		return (read_1820_temp(d));

	case DS2405:
		return (read_2405_status(d));

	case DS2407:
		return (read_2407_infob(d));

	case DS2408:
		return (read_2408_pins(d));

	case DS2413:
		return (read_2413_pins(d));

	case DS2423:
		return (read_2423_counts(d));

	case DS2450:
		return (read_2450_adc(d));

	case DS2436:
		return (read_2436_data(d));

	case DS2438:
		return (read_2438_data(d));
	}
	return (FALSE); /* 2401 are used in a peculiar way */
}

static void
fondle_each_separately()
{
	struct owdev *d;

	for (d = devices; d < end_devices; ++d)

		if (fondle(d) && debug)
			prdevice(d);
}

/* once per few seconds is good, min 750 msec */

void
ow_round()
{
	/* are there any at all ? */

	if (!sortiment)
		return;

	/* every 2436 and 2438 */

	if (any(DS2436) || any(DS2438)) {

		if (++roundcnt & 1) {
			SKIP_cmd(CONVERT_V);
			return;              /* next time, more of this */
		}
		if (any(DS2438))
			recall_every_2438(); /* both temp and a/d results */
	}
	fondle_each_separately();

	/* after initial fondling, to differentiate 1 and 2 channel 2407 */

	if (numrank == 0)
		rank_them();

	if (any(DS2450))
		start_every_2450();

	/* every 1820 and 2438 */

	if (any(DS1820) || any(DS1822) || any(DS2438))
		SKIP_cmd(CONVERT_T);

	/* every 2436 CONVERT_T (oddly not the same command as in 1820 and 2438) */

	if (any(DS2436))
		SKIP_cmd(0xD2);

	if (1)
		scan_2401();
}

/*
 * call as many times as you want, to handle disappearing and
 * reappearing devices
 */
void
ow_reinit()
{
	/* preconfigure the i/o-pin after reboot */

	if (end_devices == NULL)
		ow_reset();

	sortiment = 0;
	numrank = byterank = accrank = 0;
	end_devices = devices;
	bzero(devices, sizeof(devices));

	if (!ow_reset()) /* only place where we care about the presence pulse */
		return;

	command(SEARCH_ROM);
	scan_bus(0);

	if (any(DS2438))
		config_every_2438();
}

